// ---------------------------------------------------------------------------
// - Axo.cpp                                                                 -
// - afnix object collection - command class implementation                  -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Axo.hpp"
#include "System.hpp"
#include "Integer.hpp"
#include "Function.hpp"
#include "Axocalls.hpp"
#include "InputFile.hpp"
#include "InputMapped.hpp"
#include "Exception.hpp"
#include "OutputFile.hpp"

namespace afnix {

  // query default prompts
  static const char* DEFAULT_PROMPT1 = "(axo)";
  static const char* DEFAULT_PROMPT2 = "....>";
    
  // this function initialize the axo nameset
  static Nameset* initialize_axo (Axo* ctrl) {
    // create the axo nameset
    Nameset* axo = ctrl->mknset ("axo", ctrl->getgset ());
    // bind the axo commands
    axo->symcst ("new",          new Function (axo_new));
    axo->symcst ("open",         new Function (axo_open));
    axo->symcst ("save",         new Function (axo_save));
    axo->symcst ("info",         new Function (axo_info));
    axo->symcst ("sort",         new Function (axo_sort));
    axo->symcst ("create",       new Function (axo_create));
    axo->symcst ("import",       new Function (axo_import));
    axo->symcst ("export",       new Function (axo_export));
    axo->symcst ("select",       new Function (axo_select));
    axo->symcst ("insert",       new Function (axo_insert));
    axo->symcst ("table-info",   new Function (axo_tinfo));
    axo->symcst ("table-report", new Function (axo_tblrpt));
    return axo;
  }

  // create a default data commands

  Axo::Axo (void) {
    // initialize the commander
    p_dc  = nilp;
    p_tbl = nilp;
    if (p_term != nilp) {
      p_term->setpprompt (DEFAULT_PROMPT1);
      p_term->setsprompt (DEFAULT_PROMPT2);
    }
    // bind the axo nameset
    Object::iref (p_cset = initialize_axo (this));
  }

  // create a stream based commander

  Axo::Axo (Input* is,Output* os,Output* es) : Interp (is,os,es) {
    // initialize the query
    p_dc  = nilp;
    p_tbl = nilp;
    // bind the axo nameset
    Object::iref (p_cset = initialize_axo (this));
  }
  
  // destroy this commander

  Axo::~Axo (void) {
    if (p_cset != nilp) p_cset->clear ();
    Object::dref (p_cset);
    Object::dref (p_tbl);
    Object::dref (p_dc);
  }

  // return the class name

  String Axo::repr (void) const {
    return "Axo";
  }

  // create a new collection
  
  Collection* Axo::newc (void) {
    wrlock ();
    // clean the old collection
    Object::dref (p_dc);
    // create a new collection
    try {
      p_dc = new Collection;
      Object::iref (p_dc);
    } catch (...) {
      unlock ();
      throw;
    }
    unlock ();
    return p_dc;
  }

  // create a new collection by name
  
  Collection* Axo::newc (const String& name) {
    wrlock ();
    // clean the old collection
    Object::dref (p_dc);
    // create a new collection
    try {
      p_dc = new Collection (name);
      Object::iref (p_dc);
    } catch (...) {
      unlock ();
      throw;
    }
    unlock ();
    return p_dc;
  }

  // open an collection by name
  
  Collection* Axo::open (const String& fname) {
    wrlock ();
    // clean the old collection
    Object::dref (p_dc);
    try {
      InputMapped is (fname);
      p_dc = new Collection (&is);
      Object::iref (p_dc);
    } catch (...) {
      unlock ();
      throw;
    }
    unlock ();
    return p_dc;
  }

  // save a collection by name

  void Axo::save (const String& fname) {
    rdlock ();
    if (p_dc == nilp) {
      unlock ();
      throw Exception ("axo-error", "data collection is not open");
    }
    try {
      // try to open the file
      OutputFile os (fname);
      p_dc->write (os);
    } catch (...) {
      unlock ();
      throw;
    }
    unlock ();
  }
  
  // print some information about the collection

  void Axo::info (Output* os) const {
    rdlock ();
    if (p_dc == nilp) {
      unlock ();
      throw Exception ("axo-error", "data collection is not open");
    }
    try {
      // create a print table
      PrintTable result (2);
      long row = result.add ();
      result.setsize (0, 40);
      result.set (row, 0, "Object");
      result.set (row, 1, "Value");
      // get collection name
      String cname = p_dc->getname ();
      row = result.add ();
      result.set (row, 0, "Collection name");
      result.set (row, 1, cname);
      // get the current table
      String tname = (p_tbl == nilp) ? "none" : p_tbl->getname ();
      row = result.add ();
      result.set (row, 0, "Current table");
      result.set (row, 1, tname);
      // print the table
      if (os != nilp) result.format (*os);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // print collection table contents

  void Axo::tinfo (Output* os) const {
    rdlock ();
    if (p_dc == nilp) {
      unlock ();
      throw Exception ("axo-error", "data collection is not open");
    }
    try {
      // create a print table
      PrintTable result (3);
      long row = result.add ();
      result.set (row, 0, "Table index");
      result.set (row, 1, "Table name");
      result.set (row, 2, "Table length");
      // collect table info
      long len = p_dc->length ();
      for (long i = 0; i < len; i++) {
	Table* tbl = p_dc->get (i);
	if (tbl == nilp) continue;
	Integer ival (i);
	row = result.add ();
	result.set (row, 0, ival.tostring ());
	result.set (row, 1, tbl->getname ());
	ival = tbl->length ();
	result.set (row, 2, ival.tostring ());
      }
      if (os != nilp) result.format (*os);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // create a new table and make it the default

  Table* Axo::create (void) {
    wrlock ();
    if (p_dc == nilp) {
      unlock ();
      throw Exception ("axo-error", "data collection is not open");
    }
    // remove the old table and create a new one
    Object::dref (p_tbl);
    Object::iref (p_tbl = new Table);
    // attach it to the collection
    p_dc->add (p_tbl);
    unlock ();
    return p_tbl;
  }

  // import data into the default table

  void Axo::import (const String& fname) {
    wrlock ();
    if (p_dc == nilp) {
      unlock ();
      throw Exception ("axo-error", "data collection is not open");
    }
    if (p_tbl == nilp) {
      unlock ();
      throw Exception ("axo-error", "no current table is set");
    }
    InputFile* is = nilp;
    try {
      // try to open the file
      Object::iref (is = new InputFile (fname));
      p_tbl->import (is);
      unlock ();
    } catch (Exception& e) {
      e.setfname (fname);
      Object::dref (is);
      unlock ();
      throw;
    } catch (...) {
      Object::dref (is);
      unlock ();
      throw;
    }
  }

  // select a table by index and make it the default one

  void Axo::select (const long index) {
    wrlock ();
    if (p_dc == nilp) {
      unlock ();
      throw Exception ("axo-error", "data collection is not open");
    }
    try {
      Table* tbl = p_dc->get (index);
      Object::iref (tbl);
      Object::dref (p_tbl);
      p_tbl = tbl;
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // insert element to the current table

  void Axo::insert (Vector* elems) {
    wrlock ();
    if (p_dc == nilp) {
      unlock ();
      throw Exception ("axo-error", "data collection is not open");
    }
    if (p_tbl == nilp) {
      unlock ();
      throw Exception ("axo-error", "no current table is set");
    }
    try {
      p_tbl->adddata (elems);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // print a table content

  void Axo::tblrpt (Output* os, long max, long start, bool flag) const {
    rdlock ();
    if (p_dc == nilp) {
      unlock ();
      throw Exception ("axo-error", "data collection is not open");
    }
    if (p_tbl == nilp) {
      unlock ();
      throw Exception ("axo-error", "no current table is set");
    }
    try {
      PrintTable* table = p_tbl->convert (max, start, flag);
      table->format (*os);
      delete table;
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // sort the default table by column index and flag

  void Axo::sort (const long col, const bool mode) {
    rdlock ();
    if (p_dc == nilp) {
      unlock ();
      throw Exception ("axo-error", "data collection is not open");
    }
    if (p_tbl == nilp) {
      unlock ();
      throw Exception ("axo-error", "no current table is set");
    }
    try {
      p_tbl->sort (col, mode);
      unlock ();
    } catch (...) {
      unlock ();
      throw;
    }
  }
}
