// ---------------------------------------------------------------------------
// - Axocalls.cpp                                                            -
// - afnix object collection - command calls implementation                  -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Axo.hpp"
#include "Axocalls.hpp"
#include "OutputFile.hpp"

namespace afnix {

  // create a new collection

  Object* axo_new (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // check for 0 argument
    if (argc == 0) {
      Collection* result = axo->newc ();
      robj->post (result);
      delete argv;
      return result;
    }
    // check for 1 argument
    if (argc == 1) {
      String name = argv->getstring (0);
      Collection* result = axo->newc (name);
      robj->post (result);
      delete argv;
      return result;
    }
    delete argv;
    throw Exception ("axo-error", "invalid new arguments");
  }

  // open a collection file

  Object* axo_open (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // check for 1 argument
    if (argc == 1) {
      String fname = argv->getstring (0);
      Collection* result = axo->open (fname);
      robj->post (result);
      delete argv;
      return result;
    }
    delete argv;
    throw Exception ("axo-error", "invalid open arguments");
  }

  // save a collection in a file

  Object* axo_save (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // check for 1 argument
    if (argc == 1) {
      String fname = argv->getstring (0);
      axo->save (fname);
      delete argv;
      return nilp;
    }
    delete argv;
    throw Exception ("axo-error", "invalid save arguments");
  }

  // print some collection info

  Object* axo_info (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // check for 0 argument
    if (argc == 0) {
      axo->info (robj->getos ());
      delete argv;
      return nilp;
    }
    // check for one argument
    if (argc == 1) {
      Object* obj = argv->get (0);
      Output* os  = dynamic_cast <Output*> (obj);
      if (os == nilp) 
	throw Exception ("type-error", "invalid output stream object",
			 Object::repr (obj));
      try {
	axo->info (os);
	delete argv;
      } catch (...) {
	delete argv;
	throw;
      }
    }
    delete argv;
    throw Exception ("argument-error", "too many arguments with info");
  }

  // sort the default table

  Object* axo_sort (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // sort the table
    try {
      // check for 0 argument
      if (argc == 0) {
	axo->sort (0, true);
	delete argv;
	return nilp;
      }
      // check for 1 argument
      if (argc == 1) {
	long col = argv->getint (0);
	axo->sort (col, true);
	delete argv;
	return nilp;
      } 
      // check for 2 arguments
      if (argc == 2) {
	long col  = argv->getint  (0);
	long mode = argv->getbool (1);
	axo->sort (col, mode);
	delete argv;
	return nilp;
      }
      throw Exception ("argument-error", "too many arguments with sort");
    } catch (...) {
      delete argv;
      throw;
    }
  }

  // print the collection table info

  Object* axo_tinfo (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // check for 0 argument
    if (argc == 0) {
      axo->tinfo (robj->getos ());
      delete argv;
      return nilp;
    }
    // check for one argument
    if (argc == 1) {
      Object* obj = argv->get (0);
      Output* os  = dynamic_cast <Output*> (obj);
      if (os == nilp) 
	throw Exception ("type-error", "invalid output stream object",
			 Object::repr (obj));
      try {
	axo->tinfo (os);
	delete argv;
      } catch (...) {
	delete argv;
	throw;
      }
    }
    delete argv;
    throw Exception ("argument-error", "too many arguments with table info");
  }

  // create a new table - by name eventually - make it the default

  Object* axo_create (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // check for 0 argument
    if (argc == 0) {
      try {
	Table* tbl = axo->create ();
	delete argv;
	robj->post (tbl);
	return tbl;
      } catch (...) {
	delete argv;
	throw;
      }
    }
    // check for one argument
    if (argc == 1) {
      try {
	String name = argv->getstring (0);
	Table* tbl = axo->create ();
	tbl->setname (name);
	robj->post (tbl);
	delete argv;
	return tbl;
      } catch (...) {
	delete argv;
	throw;
      }
    }
    delete argv;
    throw Exception ("argument-error", "too many arguments with create");
  }

  // import data into the table

  Object* axo_import (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // check for one argument
    if (argc == 1) {
      try {
	String fname = argv->getstring (0);
	axo->import (fname);
	delete argv;
	return nilp;
      } catch (...) {
	delete argv;
	throw;
      }
    }
    delete argv;
    throw Exception ("argument-error", "too many arguments with import");
  }

  // export a table into a file

  Object* axo_export (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for at lest one argument
    if (argc == 0) 
      throw Exception ("argument-error", "missing export file name");
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // check for one argument
    if (argc == 1) {
      try {
	String fname = argv->getstring (0);
	OutputFile os (fname);
	axo->tblrpt (&os, 0, 0, false);
	delete argv;
	return nilp;
      } catch (...) {
	delete argv;
	throw;
      }
    }
    delete argv;
    throw Exception ("argument-error", "too many arguments with export");
  }

  // select a table - and make it the default one

  Object* axo_select (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // check for 0 argument
    if (argc == 0) {
      try {
	axo->select (0);
	delete argv;
	return nilp;
      } catch (...) {
	delete argv;
	throw;
      }
    }
    // check for one argument
    if (argc == 1) {
      try {
	long index = argv->getint (0);
	axo->select (index);
	delete argv;
	return nilp;
      } catch (...) {
	delete argv;
	throw;
      }
    }
    delete argv;
    throw Exception ("argument-error", "too many arguments with info");
  }

  // insert elements to the current table

  Object* axo_insert (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // try to add the element
    try {
      axo->insert (argv);
      delete argv;
      return nilp;
    } catch (...) {
      delete argv;
      throw;
    }
  }

  // print a report of the current table

  Object* axo_tblrpt (Runnable* robj, Nameset* nset, Cons* args) {
    // get the arguments
    Vector* argv = Vector::eval (robj, nset, args);
    long    argc = (argv == nilp) ? 0 : argv->length ();
    // get the command object
    Axo* axo = dynamic_cast <Axo*> (robj);
    // format the table
    try {
      // check for 0 argument
      if (argc == 0) {
	axo->tblrpt (axo->getos (), 0, 0, true);
	delete argv;
	return nilp;
      }
      // check for 1 argument
      if (argc == 1) {
	long max = argv->getint (0);
	axo->tblrpt (axo->getos (), max, 0, true);
	delete argv;
	return nilp;
      } 
      // check for 2 arguments
      if (argc == 2) {
	long max   = argv->getint (0);
	long start = argv->getint (1);
	axo->tblrpt (axo->getos (), max, start, true);
	delete argv;
	return nilp;
      }
      throw Exception ("argument-error", 
		       "too many arguments with table-report");
    } catch (...) {
      delete argv;
      throw;
    }
  }
}
