// ---------------------------------------------------------------------------
// - Importer.cpp                                                            -
// - afnix:odc library - importer class implementation                       -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Reader.hpp"
#include "Importer.hpp"
#include "Runnable.hpp"

namespace afnix {

  // the common supported quarks
  static const long QUARK_IMPORT = String::intern ("import");

  // -------------------------------------------------------------------------
  // - record importer class object                                          -
  // -------------------------------------------------------------------------

  // the record importer supported quarks
  static const long QUARK_SETRCD = String::intern ("set-record");
  static const long QUARK_GETRCD = String::intern ("get-record");

  // create a default record importer

  RecordImport::RecordImport (void) {
    p_ircd = nilp;
    setrcd (new Record);
  }

  // create a record importer with a record

  RecordImport::RecordImport (Record* rcd) {
    p_ircd = nilp;
    setrcd (rcd);
  }

  // destroy this record importer

  RecordImport::~RecordImport (void) {
    Object::dref (p_ircd);
  }

  // return the object name

  String RecordImport::repr (void) const {
    return "RecordImport";
  }

  // return the importation record

  Record* RecordImport::getrcd (void) const {
    rdlock ();
    Record* result = p_ircd;
    unlock ();
    return result;
  }

  // set the record to import

  void RecordImport::setrcd (Record* rcd) {
    wrlock ();
    Object::dref (p_ircd);
    Object::iref (p_ircd = rcd);
    unlock ();
  }

  // import a list of literal into the record

  void RecordImport::import (Cons* cons) {
    if (cons == nilp) return;
    wrlock ();
    try {
      // iterate in the cons cell
      while (cons != nilp) {
	// get the object and reduce it
	Object* car = cons->getcar ();
	Object* obj = (car == nilp) ? nilp : car->mini ();
	p_ircd->add (obj);
	cons = cons->getcdr ();
      }
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // import stream data into the record

  void RecordImport::import (Input* is) {
    if (is == nilp) return;
    wrlock ();
    // create a new reader
    Reader rd (is);
    wrlock ();
    Form* form = nilp;
    try {
      // parse with the reader
      while (true) {
	form = rd.parse ();
	if (form == nilp) break;
	import (form);
      }
    } catch (Exception& e) {
      if (form == nilp) {
	e.setlnum (rd.getlnum ());
      } else {
	e.setlnum (form->getlnum ());
	Object::cref (form);
      }
      unlock ();
      throw;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // create a new record importer in a generic way

  Object* RecordImport::mknew (Vector* argv) {
    // get number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new RecordImport;
    // check for 1 argument
    if (argc == 1) {
      Object* obj = argv->get (0);
      Record* rcd = dynamic_cast <Record*> (obj);
      if (rcd != nilp) {
	return new RecordImport (rcd);
      }
      throw Exception ("type-error", "invalid argument with record import");
    }
    throw Exception ("argument-error", "too many argument with record import");
  }

  // apply a record importer method with a set of arguments and a quark

  Object* RecordImport::apply (Runnable* robj, Nameset* nset, 
			       const long quark, Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETRCD) {
	rdlock ();
	Record* rcd = getrcd ();
	robj->post (rcd);
	unlock ();
	return rcd;
      }
    }
    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETRCD) {
	Object* obj = argv->get (0);
	Record* rcd = dynamic_cast <Record*> (obj);
	if (rcd == nilp) 
	  throw Exception ("type-error", "invalid object for set-record ",
			   Object::repr (obj));
	setrcd (rcd);
	return nilp;
      }
      if (quark == QUARK_IMPORT) {
	Object* obj = argv->get (0);
	// check for an input stream
	Input*   is = dynamic_cast <Input*> (obj);
	if (is != nilp) {
	  import (is);
	  return nilp;
	}
	// check for a cons cell
	Cons* cons = dynamic_cast <Cons*> (obj);
	if (cons != nilp) {
	  import (cons);
	  return nilp;
	}
	throw Exception ("type-error", "invalid object for import ",
			 Object::repr (obj));
      }
    }
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }

  // -------------------------------------------------------------------------
  // - table importer class object                                           -
  // -------------------------------------------------------------------------

  // the table importer supported quarks
  static const long QUARK_SETTBL = String::intern ("set-table");
  static const long QUARK_GETTBL = String::intern ("get-table");

  // create a default table importer

  TableImport::TableImport (void) {
    p_itbl = nilp;
    settbl (new Table);
  }

  // create a table importer with a table

  TableImport::TableImport (Table* rcd) {
    p_itbl = nilp;
    settbl (rcd);
  }

  // destroy this table importer

  TableImport::~TableImport (void) {
    Object::dref (p_itbl);
  }

  // return the object name

  String TableImport::repr (void) const {
    return "TableImport";
  }

  // return the importation table

  Table* TableImport::gettbl (void) const {
    rdlock ();
    Table* result = p_itbl;
    unlock ();
    return result;
  }

  // set the table to import

  void TableImport::settbl (Table* tbl) {
    wrlock ();
    Object::dref (p_itbl);
    Object::iref (p_itbl = tbl);
    unlock ();
  }

  // import a list of literal into the table

  void TableImport::import (Cons* cons) {
    if (cons == nilp) return;
    wrlock ();
    Record* rcd = new Record;
    try {
      // iterate in the cons cell
      while (cons != nilp) {
	// get the object and reduce it
	Object* car = cons->getcar ();
	Object* obj = (car == nilp) ? nilp : car->mini ();
	rcd->add (obj);
	cons = cons->getcdr ();
      }
      p_itbl->add (rcd);
    } catch (...) {
      Object::cref (rcd);
      unlock ();
      throw;
    }
  }

  // import stream data into the table

  void TableImport::import (Input* is) {
    if (is == nilp) return;
    wrlock ();
    // create a new reader
    Reader rd (is);
    wrlock ();
    Form* form = nilp;
    try {
      // parse with the reader
      while (true) {
	form = rd.parse ();
	if (form == nilp) break;
	import (form);
      }
    } catch (Exception& e) {
      if (form == nilp) {
	e.setlnum (rd.getlnum ());
      } else {
	e.setlnum (form->getlnum ());
	Object::cref (form);
      }
      unlock ();
      throw;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // create a new table importer in a generic way

  Object* TableImport::mknew (Vector* argv) {
    // get number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new TableImport;
    // check for 1 argument
    if (argc == 1) {
      Object* obj = argv->get (0);
      Table*  tbl = dynamic_cast <Table*> (obj);
      if (tbl != nilp) {
	return new TableImport (tbl);
      }
      throw Exception ("type-error", "invalid argument with table import");
    }
    throw Exception ("argument-error", "too many argument with table import");
  }

  // apply a table importer method with a set of arguments and a quark

  Object* TableImport::apply (Runnable* robj, Nameset* nset, 
			      const long quark, Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // dispatch 0 argument
    if (argc == 0) {
      if (quark == QUARK_GETTBL) {
	rdlock ();
	Table* tbl = gettbl ();
	robj->post (tbl);
	unlock ();
	return tbl;
      }
    }
    // dispatch 1 argument
    if (argc == 1) {
      if (quark == QUARK_SETTBL) {
	Object* obj = argv->get (0);
	Table*  tbl = dynamic_cast <Table*> (obj);
	if (tbl == nilp) 
	  throw Exception ("type-error", "invalid object for set-table ",
			   Object::repr (obj));
	settbl (tbl);
	return nilp;
      }
      if (quark == QUARK_IMPORT) {
	Object* obj = argv->get (0);
	// check for an input stream
	Input*   is = dynamic_cast <Input*> (obj);
	if (is != nilp) {
	  import (is);
	  return nilp;
	}
	// check for a cons cell
	Cons* cons = dynamic_cast <Cons*> (obj);
	if (cons != nilp) {
	  import (cons);
	  return nilp;
	}
	throw Exception ("type-error", "invalid object for import ",
			 Object::repr (obj));
      }
    }
    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }
}
