// ---------------------------------------------------------------------------
// - Hasher.cpp                                                              -
// - afnix:txt library - message hasher class implementation                 -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Md5.hpp"
#include "Sha1.hpp"
#include "Item.hpp"
#include "Vector.hpp"
#include "Hasher.hpp"

namespace afnix {

  // the hasher eval quarks
  static const long QUARK_MD5    = String::intern ("MD5");
  static const long QUARK_SHA1   = String::intern ("SHA-1");
  static const long QUARK_HASHER = String::intern ("HASHER");

  // map an enumeration item to a hasher type
  static inline Hasher::t_algo item_to_algo (const Item& item) {
    // check for an hasher item
    if (item.gettid () != QUARK_HASHER)
      throw Exception ("item-error", "item is not an hasher item");
    // map the item to the enumeration
    long quark = item.getquark ();
    if (quark == QUARK_MD5)  return Hasher::MD5;
    if (quark == QUARK_SHA1) return Hasher::SHA1;
    throw Exception ("item-error", "cannot map item to hasher algorithm");
  }

  // the hasher supported quarks
  static const long QUARK_COMPUTE = String::intern ("compute");

  // create a default hasher

  Hasher::Hasher (void) {
    d_algo = MD5;
    p_hash = new Md5;
  }

  // create a hash by type

  Hasher::Hasher (const t_algo algo) {
    d_algo = algo;
    switch (algo) {
    case MD5:
      p_hash = new Md5;
      break;
    case SHA1:
      p_hash = new Sha1;
      break;
    }
  }
      
  // destroy this hasher

  Hasher::~Hasher (void) {
    delete p_hash;
  }

  // return the class name

  String Hasher::repr (void) const {
    return "Hasher";
  }

  // compute a message hasher from a string

  String Hasher::compute (const String& msg) {
    wrlock ();
    try {
      String msgh = p_hash->compute (msg);
      unlock ();
      return msgh;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // compute a message hasher from an input stream

  String Hasher::compute (Input& is) {
    wrlock ();
    try {
      String msgh = p_hash->compute (is);
      unlock ();
      return msgh;
    } catch (...) {
      unlock ();
      throw;
    }
  }

  // evaluate a quark statically

  Object* Hasher::meval (Runnable* robj, Nameset* nset, const long quark) {
    if (quark == QUARK_MD5)  return new Item (QUARK_HASHER, QUARK_MD5);
    if (quark == QUARK_SHA1) return new Item (QUARK_HASHER, QUARK_SHA1);
    throw Exception ("eval-error", "cannot evaluate member",
		     String::qmap (quark));
  }

  // create a new hasher object

  Object* Hasher::mknew (Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();
    // check for 0 argument
    if (argc == 0) return new Hasher;
    // check for 1 argument
    if (argc == 1) {
      Object* obj = argv->get (0);
      // check for an item
      Item* iobj = dynamic_cast <Item*> (obj);
      if (iobj != nilp) return new Hasher (item_to_algo (*iobj));
      // invalid arguments
      throw Exception ("argument-error", "invalid argument for hasher");
    }
    // invalid arguments
    throw Exception ("argument-error", "too many arguments for hasher");
  }

  // apply a hasher method with a set of arguments and a quark
  
  Object* Hasher::apply (Runnable* robj, Nameset* nset, const long quark,
			 Vector* argv) {
    // get the number of arguments
    long argc = (argv == nilp) ? 0 : argv->length ();

    // check for 0 arguments
    if (argc == 1) {
      if (quark == QUARK_COMPUTE) {
	Object* obj = argv->get (0);
	// check for a literal
	Literal* lval = dynamic_cast <Literal*> (obj);
	if (lval != nilp) {
	  String msg = lval->tostring ();
	  return new String (compute (msg));
	}
	// check for a buffer
	Buffer* bval = dynamic_cast <Buffer*> (obj);
	if (bval != nilp) {
	  String msg = bval->tostring ();
	  return new String (compute (msg));
	}
	// check for an input stream
	Input* is = dynamic_cast <Input*> (obj);
	if (is != nilp) return new String (compute (*is));
	// invalid object
	throw Exception ("type-error", "invalid object for hasher compute",
			 Object::repr (obj));
      }
    }

    // call the object method
    return Object::apply (robj, nset, quark, argv);
  }

}
