// ---------------------------------------------------------------------------
// - Sha1.cpp                                                                -
// - afnix cryptography library - SHA1 class implementation                  -
// ---------------------------------------------------------------------------
// - This program is free software;  you can redistribute it  and/or  modify -
// - it provided that this copyright notice is kept intact.                  -
// -                                                                         -
// - This program  is  distributed in  the hope  that it will be useful, but -
// - without  any  warranty;  without  even   the   implied    warranty   of -
// - merchantability or fitness for a particular purpose.  In no event shall -
// - the copyright holder be liable for any  direct, indirect, incidental or -
// - special damages arising in any way out of the use of this software.     -
// ---------------------------------------------------------------------------
// - copyright (c) 1999-2005 amaury darsch                                   -
// ---------------------------------------------------------------------------

#include "Sha1.hpp"

namespace afnix {

  // -------------------------------------------------------------------------
  // - generic bytes mapping functions                                       -
  // -------------------------------------------------------------------------

  // byte to quad mapping function - the size is in bytes
  static void btoq (t_quad* dst, const t_byte* src, const long size) {
    for (long i = 0, j = 0; j < size; i++, j += 4)
      dst[i] =   ((t_quad) src[j+3])        | 
	        (((t_quad) src[j+2]) << 8)  |
	        (((t_quad) src[j+1]) << 16) | 
	        (((t_quad) src[j])   << 24);
  }

  // quad to bytes mapping functions - the size is in quad
  static void qtob (t_byte* dst, const t_quad* src, const long size) {
    for (long i = 0, j = 0; i < size; i++, j += 4) {
      dst[j+3] = (t_byte)  (src[i]        & 0xff);
      dst[j+2] = (t_byte) ((src[i] >> 8)  & 0xff);
      dst[j+1] = (t_byte) ((src[i] >> 16) & 0xff);
      dst[j]   = (t_byte) ((src[i] >> 24) & 0xff);
    }
  }

  // octa to bytes mapping function - the size is in octa
  static void otob (t_byte* dst, const t_octa* src, const long size) {
    for (long i = 0, j = 0; i < size; i++, j += 8) {
      dst[j+7] = (t_byte)  (src[i]        & 0xff);
      dst[j+6] = (t_byte) ((src[i] >> 8)  & 0xff);
      dst[j+5] = (t_byte) ((src[i] >> 16) & 0xff);
      dst[j+4] = (t_byte) ((src[i] >> 24) & 0xff);
      dst[j+3] = (t_byte) ((src[i] >> 32) & 0xff);
      dst[j+2] = (t_byte) ((src[i] >> 40) & 0xff);
      dst[j+1] = (t_byte) ((src[i] >> 48) & 0xff);
      dst[j]   = (t_byte) ((src[i] >> 56) & 0xff);
    }
  }

  // byte to character mapping function
  static char btoc (const char byte, const bool flag) {
    char data = flag ? byte & 0x0f : (byte >> 4) & 0x0f;
    if (data < 10) return '0' + data;
    return 'A' + (data - 10);
  }

  // byte to string mapping function - the size is in byte
  static String dtos (const t_byte* src, const long size) {
    String result;
    for (long i = 0; i < size; i++) {
      result = result + btoc (src[i], false);
      result = result + btoc (src[i], true);
    }
    return result;
  }

  // -------------------------------------------------------------------------
  // - logical functions                                                     -
  // -------------------------------------------------------------------------

  // quad rotate left function
  static inline t_quad rotl (t_quad x, long n) {
    return ((x << n) | (x >> (32 - n)));
  }

  // -------------------------------------------------------------------------
  // - SHA-1 definitions                                                     -
  // -------------------------------------------------------------------------

  // SHA-1 algorithm constants
  const char* SHA1_ALGO_NAME   = "SHA-1";
  const long  SHA1_BMSG_LENGTH = 64;
  const long  SHA1_BPAD_LENGTH = 64;
  const long  SHA1_BMSG_DIGEST = 20;

  // SHA-1 message padding
  const t_byte PAD[SHA1_BPAD_LENGTH] = {
    0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
  };

  // SHA-1 K0 constant ~  0 <= t <= 19
  const t_quad K0 = 0x5a827999;
  // SHA-1 K1 constant ~ 20 <= t <= 39
  const t_quad K1 = 0x6ed9eba1;
  // SHA-1 K2 constant ~ 40 <= t <= 59
  const t_quad K2 = 0x8f1bbcdc;
  // SHA-1 K3 constant ~ 60 <= t <= 79
  const t_quad K3 = 0xca62c1d6;

  // -------------------------------------------------------------------------
  // - SHA-1 processing functions                                            -
  // -------------------------------------------------------------------------
  
  // SHA-1 F0 function ~ 0 <= t <= 19
  static inline t_quad F0 (t_quad x, t_quad y, t_quad z) {
    return (x & y) | (~x & z);
  }
  // SHA-1 F1 function ~ 20 <= t <= 39
  static inline t_quad F1 (t_quad x, t_quad y, t_quad z) {
    return x ^ y ^ z;
  }
  // SHA-1 F2 function ~ 40 <= t <= 59
  static inline t_quad F2 (t_quad x, t_quad y, t_quad z) {
    return (x & y) | (x & z) | (y & z);
  }
  // SHA-1 F3 function ~ 60 <= t <= 79
  static inline t_quad F3 (t_quad x, t_quad y, t_quad z) {
    return x ^ y ^ z;
  }

  // -------------------------------------------------------------------------
  // - SHA-1 object                                                          -
  // -------------------------------------------------------------------------

  // create a default SHA-1 object

  Sha1::Sha1 (void) : Digest (SHA1_ALGO_NAME, SHA1_BMSG_LENGTH) {
    reset ();
  }
  
  // reset this SHA-1 object

  void Sha1::reset (void) {
    Digest::reset ();
    d_state[0] = 0x67452301;
    d_state[1] = 0xefcdab89;
    d_state[2] = 0x98badcfe;
    d_state[3] = 0x10325476;
    d_state[4] = 0xc3d2e1f0;
  }

  // update the SHA-1 state with the buffer data
  
  void Sha1::update (void) {
    // make sure the buffer is full
    if (getcount () != SHA1_BMSG_LENGTH) return;
    // decode a block in 16 quads
    t_quad x[16]; btoq (x, p_data, SHA1_BMSG_LENGTH);
    // prepare a message schedule
    t_quad W[80];
    for (long i = 0; i < 16; i++) W[i] = x[i];
    for (long i = 16; i < 80; i++) {
      W[i] = rotl (W[i-3]^W[i-8]^W[i-14]^W[i-16],1);
    }				    
    // initialize state values
    t_quad a = d_state[0];
    t_quad b = d_state[1];
    t_quad c = d_state[2];
    t_quad d = d_state[3];
    t_quad e = d_state[4];
    // update ~ 0 <= t <= 19
    for (long i = 0; i < 20; i++) {
      t_quad t = rotl (a,5) + F0 (b,c,d) + e + K0 + W[i];
      e = d;
      d = c;
      c = rotl (b, 30);
      b = a;
      a = t;
    }
    // update ~ 20 <= t <= 39
    for (long i = 20; i < 40; i++) {
      t_quad t = rotl (a,5) + F1 (b,c,d) + e + K1 + W[i];
      e = d;
      d = c;
      c = rotl (b, 30);
      b = a;
      a = t;
    }
    // update ~ 40 <= t <= 59
    for (long i = 40; i < 60; i++) {
      t_quad t = rotl (a,5) + F2 (b,c,d) + e + K2 + W[i];
      e = d;
      d = c;
      c = rotl (b, 30);
      b = a;
      a = t;
    }
    // update ~ 60 <= t <= 79
    for (long i = 60; i < 80; i++) {
      t_quad t = rotl (a,5) + F3 (b,c,d) + e + K3 + W[i];
      e = d;
      d = c;
      c = rotl (b, 30);
      b = a;
      a = t;
    }
    // state update
    d_state[0] += a;
    d_state[1] += b;
    d_state[2] += c;
    d_state[3] += d;
    d_state[4] += e;
    // clear the buffer count to indicate processing
    clear ();
  }
  
  // process a message by data

  void Sha1::process (const t_byte* msg, const long size) {
    long blen = size;
    while (blen != 0) {
      long step = copy (msg, blen);
      if (getcount () == SHA1_BMSG_LENGTH) {
	update ();
	msg += step;
      }
      blen -= step;
    }
  }
  
  // process a message with an input stream

  void Sha1::process (Input& is) {
    while (is.iseof () == true) {
      copy (is);
      if (getcount () == SHA1_BMSG_LENGTH) update ();
    }
  }
  
  // finish a message by padding the message
  
  void Sha1::finish (void) {
    // if the buffer length is 0 do nothing
    t_long len = getlength ();
    // how much padding do we have to do - the message must be congruent
    // to 56 modulo 64 (aka 448 bits modulo 512)
    long res = len % SHA1_BPAD_LENGTH;
    long pad = (res < 56) ? 56 - res : 120 - res;
    // process with the padding
    process (PAD, pad);
    // now add the length - this is in bits !!!
    t_byte blen[8];
    t_octa bits = (t_octa) len << 3;
    otob (blen, &bits, 1);
    process (blen, 8);
  }
  
  // format the digest
  
  String Sha1::format (void) const {
    // create the message digest
    t_byte msgb[SHA1_BMSG_DIGEST];
    qtob (msgb, d_state, 5);
    // format the string result
    String result = dtos (msgb, SHA1_BMSG_DIGEST);
    return result;
  }

  // compute a message digest from a string

  String Sha1::compute (const String& msg) {
    long    size = msg.length ();
    t_byte* data = (t_byte*) msg.tochar ();
    try {
      process (data, size);
      finish  ();
      String msgh = format ();
      reset   ();
      delete [] data;
      return msgh;
    } catch (...) {
      delete [] data;
      reset ();
      throw;
    }
  }

  // compute a message digest from an input stream

  String Sha1::compute (Input& is) {
    try {
      process (is);
      finish  ();
      String msgh = format ();
      reset   ();
      return msgh;
    } catch (...) {
      reset ();
      throw;
    }
  }
}
